<?php
//C:\workfolder\v_220\products\sources\modules\Scheduler\Logic\consumer.php

namespace Scheduler\Logic;
require_once __DIR__.'/../../../../../public/bootstrap.php';
require_once 'ReportSender.php';
require_once 'MailSender.php';
require_once 'ProcedureRunner.php';
set_time_limit(0);


class Consumer{
    //@TODO: Implementar um método que consome somente um
    const GET_PROCESSO_EXECUCAO = "SELECT *
            FROM (SELECT TP.PROCESS_PATH, TPQ.TKU_PROCESS_QUEUE_ID, TPQ.PROCESS_PARAMS, TPQ.NRORG
                FROM TKU_PROCESS_QUEUE TPQ
                LEFT JOIN TKU_PROCESS TP ON
                TPQ.TKU_PROCESS_ID=TP.TKU_PROCESS_ID
                WHERE TPQ.STATUS='EXECUCAO' ORDER BY TPQ.PRIORITY ASC)
            WHERE ROWNUM<=:P_ROWNUM";
    const UPDATE_PROCESSO = "UPDATE TKU_PROCESS_QUEUE SET STATUS='EXECUTADO' WHERE TKU_PROCESS_QUEUE_ID=:P_ID";
    const GET_SERVERS="SELECT SERVER_ADDRESS, PORT FROM TKU_PROCESS_MANAGER";

    /**
     * @var Objeto do Gearman contendo os workers disponíveis
     */
    private static $gearmanClient=null;
    private static $maxStrikes=5;
    private static $currentProcess=null;

    /**
     * @return null
     */
    public static function &getCurrentProcess()
    {
        return self::$currentProcess;
    }

    /**
     * @param null $currentProcess
     */
    public static function setCurrentProcess($currentProcess)
    {
        self::$currentProcess = $currentProcess;
    }

    /**
     * @return int
     */
    public static function getMaxStrikes()
    {
        return self::$maxStrikes;
    }

    /**
     * @param int $maxStrikes
     */
    public static function setMaxStrikes($maxStrikes)
    {
        self::$maxStrikes = $maxStrikes;
    }

    /**
     * Recupera o objeto com os workers disponíveis.
     * Se não houver nenhum disponível, o objeto será bool(false)
     * @return bool|GearmanClient
     */
    public static function getGearmanClient()
    {
        if(self::$gearmanClient===null){
            self::setGearmanClient(self::checkForGearman());
        }
        return self::$gearmanClient;
    }

    /**
     * @param GearmanClient $gearmanClient
     */
    public static function setGearmanClient($gearmanClient)
    {
        self::$gearmanClient = $gearmanClient;
    }

    /**
     * Cria um objeto GearmanClient, adicionando a ele os servidores
     * registrados em TKU_PROCESS_MANAGER, e o retorna
     * Se não houver servidores disponíveis, retorna false
     *
     * @return bool|GearmanClient
     */
    private static function checkForGearman(){

        //@TODO implementar isso no consumer
        $db = \Setforms2\Runtime\Connection::getInstance();
        $array_servidores=$db->fetchAll(Consumer::GET_SERVERS);

        //testa se há algum worker do gearman disponível
        $hasGearman=false;
        if (!empty($array_servidores)) {
            $client = new GearmanClient();
            foreach ($array_servidores as $servidor) {
                try {
                    $result = $client->addServer($servidor["SERVER_ADDRESS"], $servidor["PORT"]);
                    $hasGearman = $result || $hasGearman;
                } catch (GearmanException $e) {
//                    echo $e->getMessage();
//                    echo "O IP $servidor[SERVER_ADDRESS] nao pode ser registrado";
                }
            }
        }
        if ($hasGearman) return $client;
        else return false;
    }

    /**
     * Recupera do banco e executa os $qtdToExecute primeiros processos marcados para execução na fila
     *
     * @throws \Exception
     */
    public static function executeProcessesInQueue($qtdToExecute=5){

        $db = \Setforms2\Runtime\Connection::getInstance();

        //pega todos os processos com status EXECUCAO
        $arrayParams["P_ROWNUM"]=$qtdToExecute;
        $lista_processos=$db->fetchAll(Consumer::GET_PROCESSO_EXECUCAO, $arrayParams);
        essafucniocasonaoexoiste();

        foreach ($lista_processos as $processo){
            try{

                Consumer::setCurrentProcess($processo['TKU_PROCESS_QUEUE_ID']);
                $args["WORKLOAD"]=$processo['PROCESS_PARAMS'];
                $args["NRORG"]=$processo['NRORG'];
                //essafuncaonaoexiste();

                \Scheduler\Logic\Consumer::consume($processo["PROCESS_PATH"], $args);

                $arrayParams["P_ID"]=$processo['TKU_PROCESS_QUEUE_ID'];


                $db->execute(Consumer::UPDATE_PROCESSO,$arrayParams);
            }
            //para contornar exceptions relacionadas à banco de dados, que não devem fazer nada com o processo
            //pega-as antes de serem transformadas em ConsumerException e joga-as para cima
            catch(\Doctrine2\DBAL\Driver\OCI8\OCI8Exception $e){
                //$logger=\Setforms2\Runtime\Logger::getInstance();
                //$logger->error($e->getMessage());
                //throw $e;
                $errorMessage="Erro de DB - ".$e->getMessage();
                //\Scheduler\Logic\MailSender::sendErrorEmail($errorMessage);
                throw new \Exception($errorMessage);
            }
            catch(\Exception $e){
                //$logger=\Setforms2\Runtime\Logger::getInstance();
                //$logger->error($e->getMessage());
                $errorMessage="Erro no processo com ID ".$processo["TKU_PROCESS_QUEUE_ID"]
                    ." - ".$e->getMessage();
                throw new \Scheduler\Logic\ConsumerException($errorMessage,$processo["TKU_PROCESS_QUEUE_ID"]);
            }
        }

    }


    /**
     * Executa o processo marcado como AGENDADO de id $id.
     *
     * @param $id
     *  ID do processo a ser executado
     *
     * @throws \Exception
     */
    public static function executeProcess($id){

        if(empty($id)){
            throw new Exception('Nenhum parâmetro foi passado.');
        }
        require_once "../Entities/TkuProcessQueue.php";
        require_once "../Entities/TkuProcess.php";

        try{

            $processQueueResult=\Scheduler\Entities\TkuProcessQueue::findByTkuProcessQueueIdAndStatus($id,"AGENDADO");

            $processQueue=$processQueueResult[0];

            $processTypeResult=\Scheduler\Entities\TkuProcess::findByTkuProcessId($processQueue->getTkuProcessId());
            $processType=$processTypeResult[0]->getProcessPath();

            $args["WORKLOAD"]=$processQueue->getProcessParams();
            $args["NRORG"]=$processQueue->getNrorg();
            Consumer::setCurrentProcess($id);

            \Scheduler\Logic\Consumer::consume($processType, $args);

        }
        catch(\Doctrine2\DBAL\Driver\OCI8\OCI8Exception $e){
//            $logger=\Setforms2\Runtime\Logger::getInstance();
//            $logger->error($e->getMessage());
            $errorMessage="Erro de DB - ".$e->getMessage();
            //\Scheduler\Logic\MailSender::sendErrorEmail($errorMessage);
            throw new \Exception($errorMessage);
        }
        catch(\Exception $e){
//            $logger=\Setforms2\Runtime\Logger::getInstance();
//            $logger->error($e->getMessage());
            $errorMessage="Erro no processo com ID ".$id
                ." - ".$e->getMessage();
            //\Scheduler\Logic\MailSender::sendErrorEmail($errorMessage);
            throw new \Scheduler\Logic\ConsumerException($errorMessage, $id);
        }

    }

    /**
     * Executa o job $jobType com os argumentos em $args. Se há workers do Gearman disponíveis
     * (i.e. self::$gearmanClient!=false), utiliza-os. Se não, relega a função à doJob()
     *
     * @param $jobType
     *    Nome do trabalho a ser executado. Deve ser implementado ou pelos workers do Gearman ou por doJob()
     *
     * @param $args
     *    Array com dois índices. O índice WORKLOAD é uma string, a carga de trabalho
     *      (essa deverá ser interpretada pela implementação do trabalho, que a receberá como parâmetro)
     *    e a posição NRORG é o nº da organização para a qual consultas de banco devem se referir
     *
     * @throws \Exception
     */
    public static function consume($jobType, $args){
        $client=self::getGearmanClient();

//        try{
            if ($client){
                $client->doBackground($jobType, $args);
            }
            else{
                self::doJob($jobType, $args);
            }
//        }
//        catch(\Exception $e) {
//            throw new \Exception("Erro - " . $e->getMessage());
//        }
    }

    /**
     * Executa o trabalho descrito em $job, passando como argumento $args["WORKLOAD"] e $args["NRORG"].
     * Funciona de modo *similar* à um GearmanWorker, sendo usado quando nenhum está disponível
     *
     * @param $job
     *    O nome do trabalho a ser executado. Deve estar aqui implementado.
     * @param $args
     *    Array com dois índices. O índice WORKLOAD é uma string, a carga de trabalho
     *      (essa deverá ser interpretada pela implementação do trabalho, que a receberá como parâmetro)
     *    e a posição NRORG é o nº da organização para a qual consultas de banco devem se referir.
     *    Ambos são passados como argumentos para as funções que implementam tais trabalhos
     *
     * @throws \Exception
     */
    public static function doJob($job, $args){

        $nrorg=$args["NRORG"];
        if(empty($nrorg)){
            $nrorg=1;
        }

        $workload=$args["WORKLOAD"];

        switch ($job){
            case "send_report":
                try{
                    \Scheduler\Logic\ReportSender::mandar_relatorio($workload,$nrorg);
                }
                catch(\Exception $e){
//                    $logger=\Setforms2\Runtime\Logger::getInstance();
//                    $logger->error($e->getMessage());
                    throw new \Exception("send_report - ".$e->getMessage());
                }
                break;

            case "send_email":
                try{
                    \Scheduler\Logic\MailSender::mandar_email($workload,$nrorg); break;
                }
                catch(\Exception $e){
//                    $logger=\Setforms2\Runtime\Logger::getInstance();
//                    $logger->error($e->getMessage());
                    throw new \Exception("send_email - ".$e->getMessage());
                }
                break;

            case "run_procedure":
                try{
                    \Scheduler\Logic\ProcedureRunner::executa_procedure($workload,$nrorg);
                }
                catch(\Exception $e){
//                    $logger=\Setforms2\Runtime\Logger::getInstance();
//                    $logger->error($e->getMessage());
                    throw new \Exception("run_procedure - ".$e->getMessage());
                }
                break;
            case "run_dll":
                try{
                    \Scheduler\Logic\DLLRunner::runDLL($workload,$nrorg);
                }
                catch(\Exception $e){
//                    $logger=\Setforms2\Runtime\Logger::getInstance();
//                    $logger->error($e->getMessage());
                    throw new \Exception("run_dll - ".$e->getMessage());
                }

        }
    }

    /**
     * Marca um "strike" para o processo com o id $id, indicando que ele falhou uma vez. Se o número de "strikes"
     * ultrapassar o limite determinado por maxStrikes, marca o processo com o status ERRO, o que o inpedirá
     * de ser executado novamente.
     * @param $id
     * @return bool
     */
    public static function checkProcessFailure($id){

        $db = \Setforms2\Runtime\Connection::getInstance();

        $getStrikes="SELECT STRIKECOUNT FROM TKU_PROCESS_QUEUE WHERE TKU_PROCESS_QUEUE_ID=:P_ID";
        $arrayParams["P_ID"]=$id;
        $strikeResult=$db->fetchAll($getStrikes, $arrayParams);
        $strikeNumber=$strikeResult[0]["STRIKECOUNT"];

        $strikeNumber++;
        if ($strikeNumber>=self::getMaxStrikes()){
            Consumer::setError($id);
            return true;
        }
        else{
            Consumer::setStrikes($id, $strikeNumber);
            return false;
        }
    }

    public static function setStrikes($id, $strikeNumber){
        $db = \Setforms2\Runtime\Connection::getInstance();
        $setStrikes="UPDATE TKU_PROCESS_QUEUE SET STRIKECOUNT=:P_COUNT WHERE TKU_PROCESS_QUEUE_ID=:P_ID";
        $arrayParams["P_ID"]=$id;
        $arrayParams["P_COUNT"]=$strikeNumber;
        $db->execute($setStrikes, $arrayParams);
    }

    public static function setError($id){
        $db = \Setforms2\Runtime\Connection::getInstance();
        $arrayParams["P_ID"]=$id;
        $markasfail="UPDATE TKU_PROCESS_QUEUE SET STATUS='ERRO' WHERE TKU_PROCESS_QUEUE_ID=:P_ID";
        $db->execute($markasfail, $arrayParams);
    }

   /* public static function reviveConsumer($id){  //revives the process when dead

        if (@is_array($e = @error_get_last()) && ($e['type'] === E_ERROR || $e['type'] === E_USER_ERROR)) {
            $markasfail="UPDATE TKU_PROCESS_QUEUE SET STATUS='ERRO' WHERE TKU_PROCESS_QUEUE_ID=$id";
            $db = \Setforms2\Runtime\Connection::getInstance();
            $db->execute($markasfail);
            $file = isset($e['file']) ? $e['file'] : '';
            $line = isset($e['line']) ? $e['line'] : '';
            $msg = isset($e['message']) ? $e['message'] : '';
            $msg .= " at {$file}({$line}).";
            \Scheduler\Logic\MailSender::sendErrorEmail("Erro fatal: - ".$msg);

//            file_put_contents("log.txt", $msg."\n", FILE_APPEND);

//            session_destroy();

            if (strtoupper(substr(PHP_OS, 0, 3)) === 'WIN') {
                $WshShell = new \COM('WScript.Shell');
                $oExec = $WshShell->Run('php consumerStandBy.php', 0, false);

            } else {
                exec('service consumerStandBy restart');
            }

            exit(); //na verdade é desnecessário, já que o comando anterior faz o processo se matar, mas w/e
        }
    } */
}
class ConsumerException extends \Exception{

    private $jobId;

    public function __construct($message, $jobId, $code=0, \Exception $previous=null){
        parent::__construct($message, $code, $previous);
        $this->setJobId($jobId);
    }
    /**
     * @return mixed
     */
    public function getJobId()
    {
        return $this->jobId;
    }

    /**
     * @param mixed $jobId
     */
    public function setJobId($jobId)
    {
        $this->jobId = $jobId;
    }


}
